/*
 * $Id: Bullet.as,v 1.2 2007/04/04 13:08:57 kenta Exp $
 *
 * Copyright 2007 Kenta Cho. Some rights reserved.
 */
package abagames.la2 {
  import flash.display.BitmapData;
  import abagames.util.Vector;
  import abagames.util.Actor;

  /**
   * Enemies bullet.
   */
  public class Bullet extends Actor {
    private static const SIZE:Number = 4.0;
    private var buffer:BitmapData;
    private var player:Player;
    private var field:Field;
    private var bullets:BulletPool;
    private var particles:ParticlePool;
    private var _pos:Vector;
    private var deg:Number;
    private var speed:Number;
    private var cnt:int;
    private var shapes:Array;
    private var _tainted:Boolean;

    public function Bullet(buffer:BitmapData, player:Player, field:Field, background:Background,
                         bullets:BulletPool, particles:ParticlePool) {
      this.buffer = buffer;
      this.player = player;
      this.field = field;
      this.bullets = bullets;
      this.particles = particles;
      _pos = new Vector;
      shapes = new Array;
      var s:Shape;
      s = new Shape(buffer, field, background);
      s.addBox(0, 0, SIZE * 2, SIZE * 2, 2, 0, 220, 220);
      shapes.push(s);
      s = new Shape(buffer, field, background);
      s.addBox(0, 0, SIZE * 2, SIZE * 2, 2, 180, 180, 250);
      shapes.push(s);
    }

    public function add(x:Number, y:Number, d:Number, s:Number, t:Boolean = false):void {
      _pos.x = x;
      _pos.y = y;
      deg = d;
      speed = s;
      _tainted = t;
      cnt = 0;
    }

    public override function update():void {
      _pos.x += Math.sin(deg) * speed;
      _pos.y += Math.cos(deg) * speed;
      if (!field.isIn(_pos)) {
        remove();
        return;
      }
      if (!_tainted &&
          (field.existsTaintedCellScreen(_pos.x - SIZE, _pos.y - SIZE) ||
           field.existsTaintedCellScreen(_pos.x + SIZE, _pos.y - SIZE) ||
           field.existsTaintedCellScreen(_pos.x - SIZE, _pos.y + SIZE) ||
           field.existsTaintedCellScreen(_pos.x + SIZE, _pos.y + SIZE))) {
        addParticles(Math.PI);
        remove();
        return;
      }
      var s:Number;
      var c:int = cnt % 30;
      if (c < 15)
        s = c * (1.0 / 15);
      else
        s = 2 - c * (1.0 / 15);
      if (_tainted)
        shapes[1].drawDistorted(_pos, s);
      else
        shapes[0].drawDistorted(_pos, s);
      cnt++;
    }

    public function addParticles(od:Number = 0):void {
      for (var i:int = 0; i < 7; i++) {
        var p:Particle = Particle(particles.getInstanceForced());
        p.add(_pos.x, _pos.y, deg + od + Math.random() * 0.4 - 0.2, speed * (2 + Math.random()), 100, 200, 200);
      }
    }

    public function checkHitToPlayer(p:Vector):Boolean {
      return (_pos.dist(p) < Player.BULLET_HIT_SIZE);
    }

    public function get pos():Vector {
      return _pos;
    }

    public function get tainted():Boolean {
      return _tainted;
    }
  }
}
