/*
 * $Id: Field.as,v 1.2 2007/04/04 13:08:57 kenta Exp $
 *
 * Copyright 2007 Kenta Cho. Some rights reserved.
 */
package abagames.la2 {
  import flash.display.BitmapData;
  import flash.geom.Rectangle;
  import abagames.util.Vector;
  import abagames.util.KeyState;
  import abagames.la2.barrage.*;

  /**
   * Cells of Conway's Game of Life.
   * A game status (including a score, a remaining number of player's ships, etc.) is also managed in this class.
   */
  public class Field {
    public static const SIZE_RATIO:Number = 1.25;
    public static const SIZE_OFFSET_RATIO:Number = (SIZE_RATIO - 1) / SIZE_RATIO;
    public static const CELLS_SQUARE_SIZE:Number = 12;
    public var offset:Vector;
    private const DEFAULT_UPDATE_INTERVAL:Number = 6;
    private const DEFAULT_FIRE_INTERVAL:Number = 1;
    private const DEFAULT_FIRE_SPEED:Number = 1;
    private const DEFAULT_TURRET_INSTANCES_NUM:Number = 3;
    private var buffer:BitmapData;
    private var _size:Vector;
    private var _outerSize:Vector;
    private var cells:Array;
    private var cellsCnt:Array;
    private var cellsTaint:Array;
    private var cellsTaintNext:Array;
    private var _cellsWidth:int;
    private var _cellsHeight:int;
    private var cellRect:Rectangle;
    private var updateCnt:Number;
    private var shapes:Array;
    private var screenPos:Vector;
    private var turrets:TurretPool;
    private var bullets:BulletPool;
    private var particles:ParticlePool;
    private var scoreboards:ScoreboardPool;
    private var player:Player;
    private var keyState:KeyState;
    private var mainLoop:MainLoop;
    private var updateInterval:Number;
    private var fireInterval:Number;
    private var fireSpeed:Number;
    private var turretInstancesNum:Number;
    private var score:int;
    private var left:int;
    private var leftShapePos:Vector;
    private var gameoverCnt:int;
    private var titleCnt:int;
    private var aPressed:Boolean;
    private var lineRect:Rectangle;
    private var _taintedCellsNum:int;
    private var _taintedCellsCenterPosX:Number;
    private var barrages:Array;
    private var barrageIdx:int;
    private var barrageChangeCnt:int;

    public function Field(buffer:BitmapData) {
      this.buffer = buffer;
      var sz:Number = La2.SCREEN_WIDTH * SIZE_RATIO;
      sz = (int(sz / Background.HEIGHT_MAP_SQUARE_SIZE) + 1) * Background.HEIGHT_MAP_SQUARE_SIZE;
      _size = new Vector(sz, sz);
      _outerSize = new Vector(_size.x * 1.05, _size.y * 1.05);
      _cellsWidth = _size.x / CELLS_SQUARE_SIZE;
      _cellsHeight = _size.y / CELLS_SQUARE_SIZE;
      cells = new Array;
      cellsCnt = new Array;
      cellsTaint = new Array;
      cellsTaintNext = new Array;
      for (var cx:int = 0; cx <= _cellsWidth; cx++) {
        var cl:Array = new Array;
        var ccl:Array = new Array;
        var ctl:Array = new Array;
        var ctnl:Array = new Array;
        for (var cy:int = 0; cy <= _cellsHeight; cy++) {
          cl.push(false);
          ccl.push(false);
          ctl.push(false);
          ctnl.push(false);
        }
        cells.push(cl);
        cellsCnt.push(ccl);
        cellsTaint.push(ctl);
        cellsTaintNext.push(ctnl);
      }
      offset = new Vector;
      screenPos = new Vector;
      leftShapePos = new Vector;
      leftShapePos.y = 12;
      lineRect = new Rectangle;
    }

    public function setParams(turrets:TurretPool, bullets:BulletPool, particles:ParticlePool, scoreboards:ScoreboardPool, background:Background, player:Player, keyState:KeyState, mainLoop:MainLoop):void {
      this.turrets = turrets;
      this.bullets = bullets;
      this.particles = particles;
      this.scoreboards = scoreboards;
      this.player = player;
      this.keyState = keyState;
      this.mainLoop = mainLoop;
      shapes = new Array;
      var s:Shape;
      s = new Shape(buffer, this, background);
      s.addBox(0, 0, CELLS_SQUARE_SIZE - 2, CELLS_SQUARE_SIZE - 2, 2, 0, 250, 100);
      shapes.push(s);
      s = new Shape(buffer, this, background);
      s.addBox(0, 0, CELLS_SQUARE_SIZE - 2, CELLS_SQUARE_SIZE - 2, 2, 200, 150, 100);
      shapes.push(s);
      s = new Shape(buffer, this, background);
      s.addBox(0, 0, CELLS_SQUARE_SIZE - 2, CELLS_SQUARE_SIZE - 2, 2, 0, 100, 50);
      shapes.push(s);
      s = new Shape(buffer, this, background);
      s.addBox(0, 0, CELLS_SQUARE_SIZE - 2, CELLS_SQUARE_SIZE - 2, 2, 100, 50, 0);
      shapes.push(s);
      barrages = new Array;
      barrages.push(new Nway(bullets, particles, player));
      barrages.push(new NwayHat(bullets, particles, player));
      barrages.push(new Wave(bullets, particles, player));
      barrages.push(new Rapid(bullets, particles, player));
      barrages.push(new Random(bullets, particles, player));
      barrages.push(new Spray(bullets, particles, player));
      barrages.push(new Whip(bullets, particles, player));
      barrages.push(new Bar(bullets, particles, player));
      barrages.push(new Shotgun(bullets, particles, player));
    }

    public function startTitle():void {
      init();
      aPressed = true;
      left = 0;
      titleCnt = 0;
      turretInstancesNum = 0;
      turrets.instancesNum = 0;
      updateCnt = updateInterval = 15;
    }

    public function startInGame():void {
      init();
      score = 0;
      left = 2;
      titleCnt = -1;
      barrageIdx = -1;
      barrageChangeCnt = 0;
      appearedTurret();
    }

    private function init():void {
      gameoverCnt = -1;
      updateCnt = updateInterval = DEFAULT_UPDATE_INTERVAL;
      fireInterval = DEFAULT_FIRE_INTERVAL;
      fireSpeed = DEFAULT_FIRE_SPEED;
      turretInstancesNum = DEFAULT_TURRET_INSTANCES_NUM;
      removeAllCells();
    }

    public function resetDifficulty():void {
      updateInterval += (DEFAULT_UPDATE_INTERVAL - updateInterval) * 0.5;
      fireInterval += (DEFAULT_FIRE_INTERVAL - fireInterval) * 0.5;
      fireSpeed += (DEFAULT_FIRE_SPEED - fireSpeed) * 0.5;
      turretInstancesNum += (DEFAULT_TURRET_INSTANCES_NUM - turretInstancesNum) * 0.5;
    }

    public function update():void {
      if (titleCnt < 0)
        updateDifficulty();
      updateCells();
    }

    private function updateDifficulty():void {
      updateInterval *= 0.99995;
      if (updateInterval < 2)
        updateInterval = 2;
      fireInterval *= 0.99995;
      if (fireInterval < 0.2)
        fireInterval = 0.2;
      fireSpeed *= 1.0001;
      if (fireSpeed > 2)
        fireSpeed = 2;
      turretInstancesNum *= 1.00005;
      if (turretInstancesNum >= turrets.maxInstancesNum)
        turretInstancesNum = turrets.maxInstancesNum;
      turrets.instancesNum = int(turretInstancesNum);
    }

    private function updateCells():void {
      updateCnt--;
      if (updateCnt <= 0) {
        updateCnt = updateInterval;
        var x:int, y:int;
        for (y = 0; y < _cellsHeight; y++) {
          for (x = 0; x < _cellsWidth; x++) {
            if (x > 0 && x < _cellsWidth - 1 && y > 0 && y < _cellsHeight - 1)
              continue;
            if (Math.random() < 0.3) {
              cells[x][y] = true;
              if (Math.random() < 0.01 && turrets.hasNextInstance) {
                var ad:int = int(Math.atan2(x - _cellsWidth / 2, -y + _cellsHeight / 2) / Math.PI * 4);
                if (ad < 0)
                  ad += 8;
                var tr:Turret = Turret(turrets.getInstance());
                tr.add(x, y, ad, fireInterval, fireSpeed, barrages[barrageIdx]);
              }
            } else {
              cells[x][y] = false;
            }
            cellsTaint[x][y] = false;
          }
        }
        var c:int;
        for (y = 1; y < _cellsHeight - 1; y++) {
          for (x = 1; x < _cellsWidth - 1; x++) {
            c = 0;
            var t:Boolean = false;
            for (var oy:int = -1; oy <= 1; oy++) {
              for (var ox:int = -1; ox <= 1; ox++) {
                if (cells[x + ox][y + oy])
                  c++;
                if (cellsTaint[x + ox][y + oy])
                  t = true;
              }
            }
            cellsCnt[x][y] = c;
            cellsTaintNext[x][y] = t;
          }
        }
        for (y = 1; y < _cellsHeight - 1; y++) {
          for (x = 1; x < _cellsWidth - 1; x++) {
            c = cellsCnt[x][y];
            if (cells[x][y]) {
              if (c < 3 || c > 4)
                cells[x][y] = false;
            } else {
              if (c == 3)
                cells[x][y] = true;
            }
            cellsTaint[x][y] = (cells[x][y] && cellsTaintNext[x][y]);
          }
        }
      }
      _taintedCellsNum = 0;
      _taintedCellsCenterPosX = 0;
      for (y = 0; y < _cellsHeight; y++) {
        for (x = 0; x < _cellsWidth; x++) {
          if (cells[x][y]) {
            getScreenPos(x, y);
            var si:int = 0;
            if (cellsTaint[x][y]) {
              si++;
              _taintedCellsNum++;
              _taintedCellsCenterPosX += x;
            }
            if (x == 0 || x == _cellsWidth - 1 || y == 0 || y == _cellsHeight - 1)
              si += 2;
            if (si == 0 && updateCnt <= 1) {
              var p:Particle = Particle(particles.getInstanceForced());
              p.add(screenPos.x, screenPos.y, Math.random() * Math.PI * 2, Math.random() * 3 + 1,
                    100, 250, 100);
            }
            shapes[si].draw(screenPos);
          }
        }
      }
      if (taintedCellsNum > 0)
        _taintedCellsCenterPosX /= _taintedCellsNum;
      else
        _taintedCellsCenterPosX = -1;
    }

    public function appearedTurret():void {
      barrageChangeCnt--;
      if (barrageChangeCnt <= 0) {
        barrageChangeCnt = 8;
        var pbi:int = barrageIdx;
        barrageIdx = int(Math.random() * barrages.length);
        if (barrageIdx == pbi) {
          barrageIdx++;
          if (barrageIdx >= barrages.length)
            barrageIdx = 0;
        }
      }
    }

    public function drawStatus():void {
      Letter.drawNum(score, La2.SCREEN_WIDTH - 16, 16);
      for (var i:int = 0; i < left; i++) {
        leftShapePos.x = i * 16 + 12;
        player.drawShape(leftShapePos);
      }
      if (titleCnt >= 0) {
        titleCnt++;
        Letter.drawString("L.A.2", 100, 100, 1);
        lineRect.x = 0;
        lineRect.y = 120;
        lineRect.width = 200;
        lineRect.height = 4;
        buffer.fillRect(lineRect, 0x606060);
        lineRect.y -= 2;
        lineRect.width -= 2;
        buffer.fillRect(lineRect, 0xc0c0c0);
        if (titleCnt % 60 < 30)
          Letter.drawString("PUSH Z", 260, 340, 0.8);
        if (keyState.a) {
          if (!aPressed)
            mainLoop.startInGame();
        } else {
          aPressed = false;
        }
      }
      if (gameoverCnt >= 0) {
        gameoverCnt++;
        if (gameoverCnt > 30) {
          Letter.drawString("GAME OVER", 100, 150);
          if (keyState.a) {
            if (!aPressed) {
              mainLoop.startTitle();
              return;
            }
          } else {
            aPressed = false;
          }
          if (gameoverCnt > 450)
            mainLoop.startTitle();
        }
      }
    }

    private static const GLIDER:Array = [
      [[-1, 1], [0, 1], [1, 1], [1, 0], [0, -1]],
      [[1, -1], [1, 0], [1, 1], [0, -1], [-1, 0]],
      [[-1, -1], [0, -1], [1, -1], [-1, 0], [0, 1]],
      [[-1, -1], [-1, 0], [-1, 1], [0, 1], [1, 0]],
      ];

    public function addGlider(p:Vector, d:Number):void {
      var cx:int = int((p.x + CELLS_SQUARE_SIZE / 2) / CELLS_SQUARE_SIZE);
      var cy:int = int((p.y + CELLS_SQUARE_SIZE / 2) / CELLS_SQUARE_SIZE);
      var ad:Number = d;
      if (ad < 0)
        ad += Math.PI * 2;
      var gd:int = int(ad / Math.PI * 2);
      for (var i:int = 0; i < 5; i++) {
        var gx:int = cx + GLIDER[gd][i][0];
        var gy:int = cy + GLIDER[gd][i][1];
        if (gx >= 0 && gx < _cellsWidth && gy >= 0 && gy <= _cellsHeight) {
          cells[gx][gy] = true;
          cellsTaint[gx][gy] = true;
        }
      }
    }

    public function removeTaintedCells(scoreAdd:Boolean = true):int {
      var x:int, y:int;
      var removedCellsNum:int = 0;
      var removedTurretsNum:int = turrets.removeTainted();
      for (y = 0; y < _cellsHeight; y++)
        for (x = 0; x < _cellsWidth; x++)
          if (cells[x][y] && cellsTaint[x][y])
            removedCellsNum++;
      if (!scoreAdd)
        removedTurretsNum = 0;
      var sa:Boolean = true;
      if (removedTurretsNum <= 0 || removedCellsNum <= 0)
        sa = false;
      var bulletsRemovedNum:int = bullets.removeTainted(scoreboards, sa);
      var score:int = bulletsRemovedNum;
      var cnt:int = 0;
      for (y = 0; y < _cellsHeight; y++) {
        for (x = 0; x < _cellsWidth; x++) {
          if (cells[x][y] && cellsTaint[x][y]) {
            removeCell(x, y);
            getScreenPos(x, y);
            for (var i:int = 0; i < 17; i++) {
              var p:Particle = Particle(particles.getInstanceForced());
              p.add(screenPos.x, screenPos.y, Math.random() * Math.PI * 2, Math.random() * 6 + 1,
                    250, 50, 50);
            }
            score += removedTurretsNum;
            cnt++;
            if (score > 0) {
              if (scoreboards.hasNextInstance) {
                var sb:Scoreboard = Scoreboard(scoreboards.getInstance());
                if (cnt == removedCellsNum)
                  sb.add(screenPos.x, screenPos.y, score, cnt + bulletsRemovedNum, 60);
                else
                  sb.add(screenPos.x, screenPos.y, score, cnt + bulletsRemovedNum, 20);
              } else {
                addScore(score);
              }
            }
          }
        }
      }
      return removedCellsNum;
    }

    public function addScore(s:int):void {
      score += s;
    }

    public function destroyedPlayer():Boolean {
      removeTaintedCells(false);
      left--;
      if (left < 0)
        startGameover();
      return (left >= 0);
    }

    private function startGameover():void {
      mainLoop.fadeoutBgm();
      gameoverCnt = 0;
      aPressed = true;
    }

    private function removeAllCells():void {
      for (var y:int = 0; y < _cellsHeight; y++)
        for (var x:int = 0; x < _cellsWidth; x++)
          removeCell(x, y);
    }

    public function removeCell(x:int, y:int):void {
      cells[x][y] = false;
      cellsTaint[x][y] = false;
    }

    public function getScreenPos(x:Number, y:Number):Vector {
      screenPos.x = x * CELLS_SQUARE_SIZE + CELLS_SQUARE_SIZE / 2;
      screenPos.y = y * CELLS_SQUARE_SIZE + CELLS_SQUARE_SIZE / 2;
      return screenPos;
    }

    public function existsNotTaintedCellScreen(p:Vector):Boolean {
      var cx:int = int((p.x + CELLS_SQUARE_SIZE / 2) / CELLS_SQUARE_SIZE);
      var cy:int = int((p.y + CELLS_SQUARE_SIZE / 2) / CELLS_SQUARE_SIZE);
      if (cx < 0 || cx >= _cellsWidth || cy < 0 || cy >= _cellsHeight)
        return false;
      return cells[cx][cy] && !cellsTaint[cx][cy];
    }

    public function existsTaintedCellScreen(x:Number, y:Number):Boolean {
      var cx:int = int((x + CELLS_SQUARE_SIZE / 2) / CELLS_SQUARE_SIZE);
      var cy:int = int((y + CELLS_SQUARE_SIZE / 2) / CELLS_SQUARE_SIZE);
      if (cx < 0 || cx >= _cellsWidth || cy < 0 || cy >= _cellsHeight)
        return false;
      return cellsTaint[cx][cy];
    }

    public function existsCell(x:Number, y:Number):Boolean {
      if (x < 0 || x >= _cellsWidth || y < 0 || y >= _cellsHeight)
        return false;
      return cells[int(x)][int(y)];
    }

    public function existsTaintedCell(x:Number, y:Number):Boolean {
      if (x < 0 || x >= _cellsWidth || y < 0 || y >= _cellsHeight)
        return false;
      return cellsTaint[int(x)][int(y)];
    }

    public function isIn(pos:Vector):Boolean {
      return (pos.x >= 0 && pos.x < _size.x && pos.y >= 0 && pos.y < _size.y);
    }

    public function isInOuter(pos:Vector):Boolean {
      return (pos.x >= 0 && pos.x < _outerSize.x && pos.y >= 0 && pos.y < _outerSize.y);
    }

    public function get size():Vector {
      return _size;
    }

    public function get outerSize():Vector {
      return _outerSize;
    }

    public function get cellsWidth():Number {
      return _cellsWidth;
    }

    public function get cellsHeight():Number {
      return _cellsHeight;
    }

    public function get taintedCellsNum():int {
      return _taintedCellsNum;
    }

    public function get taintedCellsCenterPosX():Number {
      return _taintedCellsCenterPosX;
    }
  }
}
