/*
 * $Id: Shape.as,v 1.2 2007/04/04 13:08:57 kenta Exp $
 *
 * Copyright 2007 Kenta Cho. Some rights reserved.
 */
package abagames.la2 {
  import flash.display.BitmapData;
  import abagames.util.Vector;

  /**
   * Shape consists of boxes and shadows.
   */
  public class Shape {
    public static var distortion:Number = 1;
    private var buffer:BitmapData;
    private var field:Field;
    private var background:Background;
    private var boxes:Array;

    public function Shape(buffer:BitmapData, field:Field, background:Background = null) {
      this.buffer = buffer;
      this.field = field;
      this.background = background;
      boxes = new Array;
    }

    public function addBox(x:Number, y:Number, w:Number, h:Number, t:Number,
                           r:Number, g:Number, b:Number):void {
      var box:Box = new Box(buffer, field, background, x, y, w, h, t, r, g, b);
      boxes.push(box);
    }

    public function draw(p:Vector, shadow:Number = 1, withOffset:Boolean = true):void {
      for each (var box:Box in boxes)
        box.draw(p, shadow, withOffset);
    }

    public function drawTilt(p:Vector, deg:Number, shadow:Number = 1):void {
      for each (var box:Box in boxes)
        box.drawTilt(p, deg, shadow);
    }

    public function drawDistorted(p:Vector, shadow:Number = 1):void {
      for each (var box:Box in boxes)
        box.drawDistorted(p, shadow);
    }
  }
}

import flash.display.BitmapData;
import flash.geom.Rectangle;
import abagames.util.Vector;
import abagames.la2.Field;
import abagames.la2.Background;
import abagames.la2.Shape;

class Box {
  private var buffer:BitmapData;
  private var field:Field;
  private var background:Background;
  private var pos:Vector;
  private var size:Vector;
  private var depth:Number;
  private var r:int;
  private var g:int;
  private var b:int;
  private var rect:Rectangle;
  private var color:uint, depthColor:uint;
  private var tp:Vector;

  public function Box(buffer:BitmapData, field:Field, background:Background,
                      x:Number, y:Number, w:Number, h:Number, d:Number,
                      r:Number, g:Number, b:Number) {
    this.buffer = buffer;
    this.field = field;
    this.background = background;
    pos = new Vector(x - w / 2, y - h / 2);
    size = new Vector(w, h);
    rect = new Rectangle(0, 0, w, h);
    this.depth = d;
    this.r = r;
    this.g = g;
    this.b = b;
    color = this.r * 0x10000 + this.g * 0x100 + this.b;
    tp = new Vector;
  }

  public function draw(p:Vector, shadow:Number, withOffset:Boolean = true):void {
    tp.x = tp.y = 0;
    rect.x = p.x + pos.x + depth;
    rect.y = p.y + pos.y + depth;
    if (withOffset) {
      rect.x -= field.offset.x;
      rect.y -= field.offset.y;
    }
    drawRect(p, shadow);
  }

  public function drawTilt(p:Vector, deg:Number, shadow:Number):void {
    tp.x = Math.cos(deg) * pos.x - Math.sin(deg) * pos.y;
    tp.y = Math.sin(deg) * pos.x + Math.cos(deg) * pos.y;
    rect.x = p.x - tp.x - field.offset.x + depth;
    rect.y = p.y + tp.y - field.offset.y + depth;
    drawRect(p, shadow);
  }

  public function drawDistorted(p:Vector, shadow:Number):void {
    tp.x = tp.y = 0;
    rect.x = p.x + pos.x + depth;
    rect.y = p.y + pos.y + depth;
    rect.x -= field.offset.x;
    rect.y -= field.offset.y;
    rect.width = size.x * Shape.distortion;
    rect.height = size.y * Shape.distortion;
    rect.x -= (rect.width - size.x) / 2;
    rect.y -= (rect.height - size.y) / 2;
    drawRect(p, shadow);
  }

  private function drawRect(p:Vector, shadow:Number):void {
    if (shadow > 0) {
      depthColor = 
        int(this.r * 0.8 * shadow) * 0x10000 +
          int(this.g * 0.8 * shadow) * 0x100 +
            int(this.b * 0.8 * shadow);
      buffer.fillRect(rect, depthColor);
    }
    rect.x -= depth;
    rect.y -= depth;
    buffer.fillRect(rect, color);
    if (background != null)
      background.addShadow(p.x - tp.x, p.y + tp.y, size);
  }
}
