/*
 * $Id: Turret.as,v 1.2 2007/04/04 13:08:57 kenta Exp $
 *
 * Copyright 2007 Kenta Cho. Some rights reserved.
 */
package abagames.la2 {
  import flash.display.BitmapData;
  import abagames.util.Vector;
  import abagames.util.Actor;
  import abagames.la2.barrage.Barrage;
  import abagames.la2.barrage.BarrageState;

  /**
   * Turret that moves over a cell and fires bullets.
   */
  public class Turret extends Actor {
    static private var moveDir:Array = [[0, -1], [1, -1], [1, 0], [1, 1], [0, 1], [-1, 1], [-1, 0], [-1, -1]];
    static private var movePattern:Array = [
      [4, 5, 3, 6, 2, 7, 1, 0], [6, 5, 7, 4, 0, 3, 1, 2], [6, 7, 5, 0, 4, 1, 3, 2], [0, 7, 1, 6, 2, 5, 3, 4],
      [0, 1, 7, 2, 6, 3, 5, 4], [2, 1, 3, 0, 4, 7, 5, 6], [2, 3, 1, 4, 0, 5, 7, 6], [4, 3, 5, 2, 6, 1, 7, 0]];
    private var buffer:BitmapData;
    private var player:Player;
    private var field:Field;
    private var bullets:BulletPool;
    private var particles:ParticlePool;
    private var pos:Vector;
    private var deg:Number;
    private var degVel:Number;
    private var screenPos:Vector;
    private var screenTargetPos:Vector;
    private var fireCnt:int;
    private var fireIntervalRatio:Number;
    private var fireSpeedRatio:Number;
    private var movePatternIdx:int;
    private var shapes:Array;
    private var _tainted:Boolean;
    private var rank:Number;
    private var onEdge:Boolean;
    private var barrage:Barrage;
    private var barrageState:BarrageState;

    public function Turret(buffer:BitmapData, player:Player, field:Field,
                         bullets:BulletPool, particles:ParticlePool) {
      this.buffer = buffer;
      this.player = player;
      this.field = field;
      this.bullets = bullets;
      this.particles = particles;
      pos = new Vector;
      screenPos = new Vector;
      screenTargetPos = new Vector;
      shapes = new Array;
      var s:Shape;
      s = new Shape(buffer, field);
      s.addBox(-5, -5, 6, 6, 1, 0, 80, 200);
      s.addBox(5, -5, 6, 6, 1, 0, 80, 200);
      s.addBox(-5, 5, 6, 6, 1, 0, 80, 200);
      s.addBox(5, 5, 6, 6, 1, 0, 80, 200);
      s.addBox(0, 0, 8, 8, 3, 50, 150, 250);
      shapes.push(s);
      s = new Shape(buffer, field);
      s.addBox(-5, -5, 6, 6, 1, 120, 0, 200);
      s.addBox(5, -5, 6, 6, 1, 120, 0, 200);
      s.addBox(-5, 5, 6, 6, 1, 120, 0, 200);
      s.addBox(5, 5, 6, 6, 1, 120, 0, 200);
      s.addBox(0, 0, 8, 8, 3, 150, 50, 250);
      shapes.push(s);
      barrageState = new BarrageState;
    }

    public function add(x:int, y:int, mpi:int, fireInterval:Number, fireSpeed:Number, barrage:Barrage):void {
      pos.x = x;
      pos.y = y;
      deg = degVel = 0;
      var sp:Vector = field.getScreenPos(pos.x, pos.y);
      screenPos.x = screenTargetPos.x = sp.x;
      screenPos.y = screenTargetPos.x = sp.y;
      movePatternIdx = mpi;
      barrageState.init(fireInterval, fireSpeed);
      this.barrage = barrage;
      _tainted = false;
      rank = 0;
      onEdge = true;
    }

    public override function update():void {
      if (!field.existsCell(pos.x, pos.y)) {
        var mv:Boolean = false;
        for (var i:int = 0; i < 8; i++) {
          var md:int = movePattern[movePatternIdx][i];
          var px:int = pos.x + moveDir[md][0];
          var py:int = pos.y + moveDir[md][1];
          if (px >= 1 && px <= field.cellsWidth - 2 && py >= 1 && py <= field.cellsWidth - 2 &&
              field.existsCell(px, py)) {
            pos.x = px;
            pos.y = py;
            var sp:Vector = field.getScreenPos(pos.x, pos.y);
            screenTargetPos.x = sp.x;
            screenTargetPos.y = sp.y;
            mv = true;
            if (onEdge) {
              field.appearedTurret();
              onEdge = false;
            }
            break;
          }
        }
        if (!mv) {
          if (onEdge) {
            remove();
            return;
          }
          degVel *= 0.96;
          barrageState.onCell = false;
        }
      } else {
        if (!tainted && field.existsTaintedCell(pos.x, pos.y))
          _tainted = barrageState.tainted = true;
        degVel += (0.1 - degVel) * 0.08;
        barrageState.onCell = true;
      }
      if (!onEdge) {
        screenPos.x += (screenTargetPos.x - screenPos.x) * 0.1;
        screenPos.y += (screenTargetPos.y - screenPos.y) * 0.1;
        deg += degVel;
        barrage.update(screenPos, barrageState);
        if (_tainted)
          shapes[1].drawTilt(screenPos, deg);
        else
          shapes[0].drawTilt(screenPos, deg);
      }
    }

    public function destroy():void {
      for (var k:int = 0; k < 23; k++) {
        var pt:Particle = Particle(particles.getInstanceForced());
        pt.add(screenPos.x, screenPos.y, Math.random() * Math.PI * 2,
               Math.random() * 7 + 3, 100, 150, 250);
      }
      remove();
    }

    public function get cellX():int {
      return int(pos.x);
    }

    public function get cellY():int {
      return int(pos.y);
    }

    public function get tainted():Boolean {
      return _tainted;
    }
  }
}
