/*
 * $Id: ActorPool.as,v 1.2 2007/04/04 13:08:58 kenta Exp $
 *
 * Copyright 2007 Kenta Cho. Some rights reserved.
 */
package abagames.util {
  import abagames.util.Actor;

  /**
   * An object pool for actors.
   */
  public class ActorPool {
    protected var actors:Array;
    protected var actorNum:int;
    protected var actorIdx:int;
    protected var _maxInstancesNum:int;
    protected var _instancesNum:int;

    public function ActorPool(n:int) {
      actors = new Array;
      for (var i:int = 0; i < n; i++) {
        var a:Actor = createActor();
        a.setActorPool(this);
        a.index = i;
        a.removed = false;
        actors.push(a);
      }
      _maxInstancesNum = _instancesNum = n;
      clear();
    }

    public function clear():void {
      actorNum = 0;
      actorIdx = -1;
    }

    public function getInstance():Actor {
      if (actorNum >= _instancesNum)
        return null;
      actorNum++;
      return actors[actorNum - 1];
    }

    public function get hasNextInstance():Boolean {
      return (actorNum < _instancesNum);
    }

    public function getInstanceForced():Actor {
      var a:Actor = getInstance();
      if (a != null)
        return a;
      actorIdx++;
      if (actorIdx >= actors.length)
        actorIdx = 0;
      return actors[actorIdx];
    }

    public function remove(idx:int):void {
      actorNum--;
      var ta:Actor = actors[idx];
      actors[idx] = actors[actorNum];
      actors[idx].index = idx;
      actors[actorNum] = ta;
      actors[actorNum].index = actorNum;
      actors[actorNum].removed = true;
    }

    public function update():void {
      for (var i:int = 0; i < actorNum; i++) {
        var a:Actor = actors[i];
        a.update();
        if (a.removed) {
          a.removed = false;
          i--;
        }
      }
    }

    protected function createActor():Actor {
      return null;
    }

    public function existsActor(idx:int):Boolean {
      return (idx < actorNum);
    }

    public function get maxInstancesNum():int {
      return _maxInstancesNum;
    }

    public function get instancesNum():int {
      return _instancesNum;
    }

    public function set instancesNum(v:int):void {
      if (v > _maxInstancesNum)
        _instancesNum = _maxInstancesNum;
      else
        _instancesNum = v;
    }
  }
}
